@@ -392,6 +392,21 @@ class GetIdsFromAliasesCall(FunctionCall):
             "out": self.out
         }
 
+@dataclass
+class GetIdsFromDagCall(FunctionCall):
+    """Call to get_ids_from_dag function."""
+
+    def to_json(self) -> dict:
+        dag = self.args['dag']
+        if isinstance(dag, FunctionCall):
+            dag = {"$ref": dag.out}
+
+        return {
+            "fn": self.fn,
+            "args": {"dag": dag},
+            "out": self.out
+        }
+
 
 @dataclass
 class PropagateCall(FunctionCall):
@@ -682,7 +697,8 @@ class OatClient:
     def set_primitive(
         self,
         id: Union[str, FunctionCall],
-        bound: complex = 1j
+        bound: complex = 1j,
+        execute_now: bool = False
     ) -> SetPrimitiveCall:
         """
         Create a primitive variable.
@@ -693,19 +709,23 @@ class OatClient:
 
         Returns:
             SetPrimitiveCall that can be referenced by other calls
+            Or if execute_now=True, returns execution result dict
         """
         call = SetPrimitiveCall(
             fn="set_primitive",
             args={"id": id, "bound": bound}
         )
         self._buffer.append(call)
+        if execute_now:
+            return self.execute([call.out], clear_buffer=True)
         return call
 
     def set_gelineq(
         self,
         coefficients: Union[Dict[Union[str, FunctionCall], int], FunctionCall],
         bias: Union[int, FunctionCall],
-        alias: Optional[str] = None
+        alias: Optional[str] = None,
+        execute_now: bool = False
     ) -> SetGelineqCall:
         """
         Create a linear inequality constraint.
@@ -716,19 +736,23 @@ class OatClient:
 
         Returns:
             SetGelineqCall
+            Or if execute_now=True, returns execution result dict
         """
         call = SetGelineqCall(
             fn="set_gelineq",
             args={"coefficients": coefficients, "bias": bias, "alias": alias}
         )
         self._buffer.append(call)
+        if execute_now:
+            return self.execute([call.out], clear_buffer=True)
         return call
 
     def set_atleast(
         self,
         references: Union[List[Union[str, FunctionCall]], FunctionCall],
         value: Union[int, FunctionCall],
-        alias: Optional[str] = None
+        alias: Optional[str] = None,
+        execute_now: bool = False
     ) -> SetThresholdOperatorCall:
         """Create an 'at least' constraint."""
         args = {"references": references, "value": value}
@@ -736,13 +760,16 @@ class OatClient:
             args["alias"] = alias
         call = SetThresholdOperatorCall(fn="set_atleast", args=args)
         self._buffer.append(call)
+        if execute_now:
+            return self.execute([call.out], clear_buffer=True)
         return call
 
     def set_atmost(
         self,
         references: Union[List[Union[str, FunctionCall]], FunctionCall],
         value: Union[int, FunctionCall],
-        alias: Optional[str] = None
+        alias: Optional[str] = None,
+        execute_now: bool = False
     ) -> SetThresholdOperatorCall:
         """Create an 'at most' constraint."""
         args = {"references": references, "value": value}
@@ -750,13 +777,16 @@ class OatClient:
             args["alias"] = alias
         call = SetThresholdOperatorCall(fn="set_atmost", args=args)
         self._buffer.append(call)
+        if execute_now:
+            return self.execute([call.out], clear_buffer=True)
         return call
 
     def set_equal(
         self,
         references: Union[List[Union[str, FunctionCall]], FunctionCall],
         value: Union[int, FunctionCall],
-        alias: Optional[str] = None
+        alias: Optional[str] = None,
+        execute_now: bool = False
     ) -> SetThresholdOperatorCall:
         """Create an 'equal' constraint."""
         args = {"references": references, "value": value}
@@ -764,12 +794,15 @@ class OatClient:
             args["alias"] = alias
         call = SetThresholdOperatorCall(fn="set_equal", args=args)
         self._buffer.append(call)
+        if execute_now:
+            return self.execute([call.out], clear_buffer=True)
         return call
 
     def set_and(
         self,
         references: Union[List[Union[str, FunctionCall]], FunctionCall],
-        alias: Optional[str] = None
+        alias: Optional[str] = None,
+        execute_now: bool = False
     ) -> SetLogicalOperatorCall:
         """Create an AND constraint."""
         args = {"references": references}
@@ -777,12 +810,15 @@ class OatClient:
             args["alias"] = alias
         call = SetLogicalOperatorCall(fn="set_and", args=args)
         self._buffer.append(call)
+        if execute_now:
+            return self.execute([call.out], clear_buffer=True)
         return call
 
     def set_or(
         self,
         references: Union[List[Union[str, FunctionCall]], FunctionCall],
-        alias: Optional[str] = None
+        alias: Optional[str] = None,
+        execute_now: bool = False
     ) -> SetLogicalOperatorCall:
         """Create an OR constraint."""
         args = {"references": references}
@@ -790,12 +826,15 @@ class OatClient:
             args["alias"] = alias
         call = SetLogicalOperatorCall(fn="set_or", args=args)
         self._buffer.append(call)
+        if execute_now:
+            return self.execute([call.out], clear_buffer=True)
         return call
 
     def set_not(
         self,
         references: Union[List[Union[str, FunctionCall]], FunctionCall],
-        alias: Optional[str] = None
+        alias: Optional[str] = None,
+        execute_now: bool = False
     ) -> SetLogicalOperatorCall:
         """Create a NOT constraint."""
         args = {"references": references}
@@ -803,12 +842,15 @@ class OatClient:
             args["alias"] = alias
         call = SetLogicalOperatorCall(fn="set_not", args=args)
         self._buffer.append(call)
+        if execute_now:
+            return self.execute([call.out], clear_buffer=True)
         return call
 
     def set_xor(
         self,
         references: Union[List[Union[str, FunctionCall]], FunctionCall],
-        alias: Optional[str] = None
+        alias: Optional[str] = None,
+        execute_now: bool = False
     ) -> SetLogicalOperatorCall:
         """Create an XOR constraint."""
         args = {"references": references}
@@ -816,13 +858,16 @@ class OatClient:
             args["alias"] = alias
         call = SetLogicalOperatorCall(fn="set_xor", args=args)
         self._buffer.append(call)
+        if execute_now:
+            return self.execute([call.out], clear_buffer=True)
         return call
 
     def set_imply(
         self,
         lhs: Union[str, FunctionCall],
         rhs: Union[str, FunctionCall],
-        alias: Optional[str] = None
+        alias: Optional[str] = None,
+        execute_now: bool = False
     ) -> SetBinaryOperatorCall:
         """Create an IMPLY constraint (lhs => rhs)."""
         args = {"lhs": lhs, "rhs": rhs}
@@ -830,13 +875,16 @@ class OatClient:
             args["alias"] = alias
         call = SetBinaryOperatorCall(fn="set_imply", args=args)
         self._buffer.append(call)
+        if execute_now:
+            return self.execute([call.out], clear_buffer=True)
         return call
 
     def set_equiv(
         self,
         lhs: Union[str, FunctionCall],
         rhs: Union[str, FunctionCall],
-        alias: Optional[str] = None
+        alias: Optional[str] = None,
+        execute_now: bool = False
     ) -> SetBinaryOperatorCall:
         """Create an EQUIV constraint (lhs <=> rhs)."""
         args = {"lhs": lhs, "rhs": rhs}
@@ -844,11 +892,14 @@ class OatClient:
             args["alias"] = alias
         call = SetBinaryOperatorCall(fn="set_equiv", args=args)
         self._buffer.append(call)
+        if execute_now:
+            return self.execute([call.out], clear_buffer=True)
         return call
 
     def get_id_from_alias(
         self,
-        alias: Union[str, FunctionCall]
+        alias: Union[str, FunctionCall],
+        execute_now: bool = False
     ) -> GetIdFromAliasCall:
         """Get node ID from alias."""
         call = GetIdFromAliasCall(
@@ -856,11 +907,14 @@ class OatClient:
             args={"alias": alias}
         )
         self._buffer.append(call)
+        if execute_now:
+            return self.execute([call.out], clear_buffer=True)
         return call
 
     def sub(
         self,
-        root: Union[str, FunctionCall]
+        root: Union[str, FunctionCall],
+        execute_now: bool = False
     ) -> SubCall:
         """Extract a sub-DAG starting from root."""
         call = SubCall(
@@ -868,6 +922,8 @@ class OatClient:
             args={"root": root}
         )
         self._buffer.append(call)
+        if execute_now:
+            return self.execute([call.out], clear_buffer=True)
         return call
 
     def solve(
@@ -875,7 +931,8 @@ class OatClient:
         dag: Union[dict, FunctionCall],
         objective: Optional[List[Dict[str, Any]]] = None,
         assume: Optional[List[Dict[str, Any]]] = None,
-        maximize: bool = True
+        maximize: bool = True,
+        execute_now: bool = False
     ) -> SolveCall:
         """
         Solve an optimization problem.
@@ -899,26 +956,32 @@ class OatClient:
             }
         )
         self._buffer.append(call)
+        if execute_now:
+            return self.execute([call.out], clear_buffer=True)
         return call
 
     def set_property(
         self,
         id: Union[str, FunctionCall],
         property: str,
-        value: Any
+        value: Any,
+        execute_now: bool = False
     ) -> SetPropertyCall:
         """Set a property value on a node."""
         call = SetPropertyCall(
             fn="set_property",
             args={"id": id, "property": property, "value": value}
         )
-        self._buffer.append(call)
+        self._buffer.append(call),
+        if execute_now:
+            return self.execute([call.out], clear_buffer=True)
         return call
 
     def set_primitives(
         self,
         ids: Union[List[Union[str, FunctionCall]], FunctionCall],
-        bound: complex = 1j
+        bound: complex = 1j,
+        execute_now: bool = False
     ) -> SetPrimitivesCall:
         """Create multiple primitive variables with the same bound."""
         call = SetPrimitivesCall(
@@ -926,11 +989,14 @@ class OatClient:
             args={"ids": ids, "bound": bound}
         )
         self._buffer.append(call)
+        if execute_now:
+            return self.execute([call.out], clear_buffer=True)
         return call
 
     def get_node_ids(
         self,
-        dag: Union[dict, FunctionCall]
+        dag: Union[dict, FunctionCall],
+        execute_now: bool = False
     ) -> GetNodeIdsCall:
         """Get all node IDs from a DAG."""
         call = GetNodeIdsCall(
@@ -938,11 +1004,14 @@ class OatClient:
             args={"dag": dag}
         )
         self._buffer.append(call)
+        if execute_now:
+            return self.execute([call.out], clear_buffer=True)
         return call
 
     def get_nodes(
         self,
-        ids: Union[List[Union[str, FunctionCall]], FunctionCall]
+        ids: Union[List[Union[str, FunctionCall]], FunctionCall],
+        execute_now: bool = False
     ) -> GetNodesCall:
         """Get multiple nodes by their IDs."""
         call = GetNodesCall(
@@ -950,11 +1019,14 @@ class OatClient:
             args={"ids": ids}
         )
         self._buffer.append(call)
+        if execute_now:
+            return self.execute([call.out], clear_buffer=True)
         return call
 
     def get_node(
         self,
-        id: Union[str, FunctionCall]
+        id: Union[str, FunctionCall],
+        execute_now: bool = False
     ) -> GetNodeCall:
         """Get a single node by its ID."""
         call = GetNodeCall(
@@ -962,11 +1034,14 @@ class OatClient:
             args={"id": id}
         )
         self._buffer.append(call)
+        if execute_now:
+            return self.execute([call.out], clear_buffer=True)
         return call
 
     def get_property_values(
         self,
-        property: str
+        property: str,
+        execute_now: bool = False
     ) -> GetPropertyValuesCall:
         """Get all nodes that have a specific property and their values."""
         call = GetPropertyValuesCall(
@@ -974,11 +1049,14 @@ class OatClient:
             args={"property": property}
         )
         self._buffer.append(call)
+        if execute_now:
+            return self.execute([call.out], clear_buffer=True)
         return call
 
     def get_alias(
         self,
-        id: Union[str, FunctionCall]
+        id: Union[str, FunctionCall],
+        execute_now: bool = False
     ) -> GetAliasCall:
         """Get alias for a given ID."""
         call = GetAliasCall(
@@ -986,11 +1064,14 @@ class OatClient:
             args={"id": id}
         )
         self._buffer.append(call)
+        if execute_now:
+            return self.execute([call.out], clear_buffer=True)
         return call
 
     def get_aliases_from_id(
         self,
-        id: Union[str, FunctionCall]
+        id: Union[str, FunctionCall],
+        execute_now: bool = False
     ) -> GetAliasesFromIdCall:
         """Get all aliases for a given ID."""
         call = GetAliasesFromIdCall(
@@ -998,11 +1079,14 @@ class OatClient:
             args={"id": id}
         )
         self._buffer.append(call)
+        if execute_now:
+            return self.execute([call.out], clear_buffer=True)
         return call
 
     def get_ids_from_aliases(
         self,
-        aliases: Union[List[str], FunctionCall]
+        aliases: Union[List[str], FunctionCall],
+        execute_now: bool = False
     ) -> GetIdsFromAliasesCall:
         """Get IDs for multiple aliases."""
         call = GetIdsFromAliasesCall(
@@ -1010,11 +1094,29 @@ class OatClient:
             args={"aliases": aliases}
         )
         self._buffer.append(call)
+        if execute_now:
+            return self.execute([call.out], clear_buffer=True)
+        return call
+    
+    def get_ids_from_dag(
+        self,
+        dag: Union[dict, FunctionCall],
+        execute_now: bool = False
+    ) -> GetIdsFromDagCall:
+        """Get all node IDs from a DAG."""
+        call = GetIdsFromDagCall(
+            fn="get_ids_from_dag",
+            args={"dag": dag}
+        )
+        self._buffer.append(call)
+        if execute_now:
+            return self.execute([call.out], clear_buffer=True)
         return call
 
     def propagate(
         self,
-        assignments: Union[List[Dict[str, Any]], FunctionCall]
+        assignments: Union[List[Dict[str, Any]], FunctionCall],
+        execute_now: bool = False
     ) -> PropagateCall:
         """Propagate constraints given initial assignments."""
         call = PropagateCall(
@@ -1022,11 +1124,14 @@ class OatClient:
             args={"assignments": assignments}
         )
         self._buffer.append(call)
+        if execute_now:
+            return self.execute([call.out], clear_buffer=True)
         return call
 
     def propagate_many(
         self,
-        many_assignments: Union[List[List[Dict[str, Any]]], FunctionCall]
+        many_assignments: Union[List[List[Dict[str, Any]]], FunctionCall],
+        execute_now: bool = False
     ) -> PropagateManyCall:
         """Propagate constraints given multiple sets of initial assignments."""
         call = PropagateManyCall(
@@ -1034,11 +1139,14 @@ class OatClient:
             args={"many_assignments": many_assignments}
         )
         self._buffer.append(call)
+        if execute_now:
+            return self.execute([call.out], clear_buffer=True)
         return call
 
     def sub_many(
         self,
-        roots: Union[List[Union[str, FunctionCall]], FunctionCall]
+        roots: Union[List[Union[str, FunctionCall]], FunctionCall],
+        execute_now: bool = False
     ) -> SubManyCall:
         """Extract sub-DAGs from multiple root nodes."""
         call = SubManyCall(
@@ -1046,6 +1154,8 @@ class OatClient:
             args={"roots": roots}
         )
         self._buffer.append(call)
+        if execute_now:
+            return self.execute([call.out], clear_buffer=True)
         return call
 
     def solve_many(
@@ -1053,7 +1163,8 @@ class OatClient:
         dag: Union[dict, FunctionCall],
         objectives: Optional[List[List[Dict[str, Any]]]] = None,
         assume: Optional[List[Dict[str, Any]]] = None,
-        maximize: bool = True
+        maximize: bool = True,
+        execute_now: bool = False
     ) -> SolveManyCall:
         """
         Solve multiple optimization problems on the same DAG.
@@ -1077,11 +1188,14 @@ class OatClient:
             }
         )
         self._buffer.append(call)
+        if execute_now:
+            return self.execute([call.out], clear_buffer=True)
         return call
 
     def validate(
         self,
-        dag: Union[dict, FunctionCall]
+        dag: Union[dict, FunctionCall],
+        execute_now: bool = False
     ) -> ValidateCall:
         """Validate a DAG structure."""
         call = ValidateCall(
@@ -1089,11 +1203,14 @@ class OatClient:
             args={"dag": dag}
         )
         self._buffer.append(call)
+        if execute_now:
+            return self.execute([call.out], clear_buffer=True)
         return call
 
     def ranks(
         self,
-        dag: Union[dict, FunctionCall]
+        dag: Union[dict, FunctionCall],
+        execute_now: bool = False
     ) -> RanksCall:
         """Compute topological ranks for nodes in a DAG."""
         call = RanksCall(
@@ -1101,11 +1218,14 @@ class OatClient:
             args={"dag": dag}
         )
         self._buffer.append(call)
+        if execute_now:
+            return self.execute([call.out], clear_buffer=True)
         return call
 
     def delete_node(
         self,
-        id: Union[str, FunctionCall]
+        id: Union[str, FunctionCall],
+        execute_now: bool = False
     ) -> DeleteNodeCall:
         """Delete a node by its ID."""
         call = DeleteNodeCall(
@@ -1113,11 +1233,14 @@ class OatClient:
             args={"id": id}
         )
         self._buffer.append(call)
+        if execute_now:
+            return self.execute([call.out], clear_buffer=True)
         return call
 
     def delete_sub(
         self,
-        roots: Union[List[Union[str, FunctionCall]], FunctionCall]
+        roots: Union[List[Union[str, FunctionCall]], FunctionCall],
+        execute_now: bool = False
     ) -> DeleteSubCall:
         """Delete sub-DAGs from the given roots."""
         call = DeleteSubCall(
@@ -1125,6 +1248,8 @@ class OatClient:
             args={"roots": roots}
         )
         self._buffer.append(call)
+        if execute_now:
+            return self.execute([call.out], clear_buffer=True)
         return call
 
     def execute(
@@ -1178,10 +1303,6 @@ class OatClient:
             if clear_buffer:
                 self.clear_buffer()
 
-            # If outputs were specified, return dict mapping output names to values
-            if outputs:
-                return dict(zip(outputs, results))
-
             # Otherwise return the raw results
             return results
 
